/*
 * Copyright 2012 Martin Skurla
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.crazyjavahacking.astvisualizer.impl.javac;


import java.util.HashSet;
import java.util.Set;

import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;

import org.crazyjavahacking.astvisualizer.api.model.NodeModelProperty;

import com.sun.source.tree.BlockTree;
import com.sun.source.tree.ImportTree;
import com.sun.source.tree.Tree;


/**
 * NodeModel properties extractor.
 *
 * @author Martin Skurla (crazyjavahacking@gmail.com)
 */
final class JavacNodeModelPropertiesExtractor {

    //-----------------------------------------------------------------------------------------------------------------
    // Constructors.
    //-----------------------------------------------------------------------------------------------------------------
    private JavacNodeModelPropertiesExtractor() {}


    //-----------------------------------------------------------------------------------------------------------------
    // Public Methods.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Extracts NodeModel properties from given com.sun.source.tree.Tree instance and mapped element. Note that element
     * will be null for all Trees that are not represented by Element.
     *
     * @param tree    tree instance
     * @param element element
     *
     * @return extracted NodeModel properties
     */
    public static Set<NodeModelProperty> extractFromTree(Tree tree, Element element) {
        Set<NodeModelProperty> properties = new HashSet<NodeModelProperty>();

        if (tree.getKind() == Tree.Kind.BLOCK) {
            processBlockTree((BlockTree) tree, properties);
        }
        else if (tree.getKind() == Tree.Kind.IMPORT) {
            processImportTree((ImportTree) tree, properties);
        }
        else if (tree.getKind() == Tree.Kind.VARIABLE) {
            processVariableTree(element, properties);
        }

        return properties;
    }

    //-----------------------------------------------------------------------------------------------------------------
    // Private Helpers.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Processes block tree.
     *
     * @param blockTree  block tree
     * @param properties properties
     */
    private static void processBlockTree(BlockTree blockTree, Set<NodeModelProperty> properties) {
        if (blockTree.isStatic()) {
            properties.add(NodeModelProperty.STATIC);
        }
    }

    /**
     * Processes import tree.
     *
     * @param importTree import tree
     * @param properties properties
     */
    private static void processImportTree(ImportTree importTree, Set<NodeModelProperty> properties) {
        if (importTree.isStatic()) {
            properties.add(NodeModelProperty.STATIC);
        }
    }

    /**
     * Processes variable tree.
     *
     * @param element    element
     * @param properties properties
     */
    private static void processVariableTree(Element element, Set<NodeModelProperty> properties) {
        if (element.getKind() == ElementKind.FIELD) {
            properties.add(NodeModelProperty.FIELD);
        }
        if (element.getKind() == ElementKind.LOCAL_VARIABLE) {
            properties.add(NodeModelProperty.LOCAL_VARIABLE);
        }
        if (element.getKind() == ElementKind.PARAMETER) {
            properties.add(NodeModelProperty.PARAMETER);
        }
    }
}
