/*
 * Copyright 2012 Martin Skurla
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.crazyjavahacking.astvisualizer.impl.javac;


import java.util.Collections;
import java.util.Set;

import javax.lang.model.element.Element;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.Name;
import javax.lang.model.type.TypeKind;

import org.crazyjavahacking.astvisualizer.api.model.NodeModel;
import org.crazyjavahacking.astvisualizer.api.model.NodeModelProperty;
import org.crazyjavahacking.astvisualizer.api.model.PositionWithinSourceCode;

import com.sun.source.tree.BlockTree;
import com.sun.source.tree.Tree;


/**
 * Node model factory.
 *
 * @author Martin Skurla (crazyjavahacking@gmail.com)
 */
final class NodeModelFactory {

    //-----------------------------------------------------------------------------------------------------------------
    // Constants.
    //-----------------------------------------------------------------------------------------------------------------
    private static final String NAME_TREE_KIND      = "Name";
    private static final String MODIFIER_TREE_KIND  = "Modifier";
    private static final String TYPE_KIND_TREE_KIND = "TypeKind";


    //-----------------------------------------------------------------------------------------------------------------
    // Constructors.
    //-----------------------------------------------------------------------------------------------------------------
    private NodeModelFactory() {}


    //-----------------------------------------------------------------------------------------------------------------
    // Public Methods.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Creates NodeModel from com.sun.source.tree.Tree type. Tree could have properties and could be leafs (however in
     * most cases they are not). Note that element will be null for all Trees that are not represented by Element.
     *
     * @param depth              node depth
     * @param order              node order
     * @param textRepresentation textual representation
     * @param javadoc            javadoc
     * @param tree               tree
     * @param element            element
     * @param sourcePosition     source position
     *
     * @return NodeModel
     *
     * @see NodeModelFactory#determineIfTreeIsLeaf(com.sun.source.tree.Tree)
     */
    public static NodeModel fromComSourceSunTreeTree(int depth, int order, String textRepresentation, String javadoc,
            Tree tree, Element element, PositionWithinSourceCode sourcePosition) {

        boolean isLeaf = determineIfTreeIsLeaf(tree);
        String treeKind = tree.getKind().toString();

        Set<NodeModelProperty> nodeModelProperties = JavacNodeModelPropertiesExtractor.extractFromTree(tree, element);

        return createNodeModel(depth, order, isLeaf, treeKind, textRepresentation, javadoc, sourcePosition,
                                                                                            nodeModelProperties);
    }

    /**
     * Creates NodeModel from javax.lang.model.element.Name type. Name does not have properties and is always leaf.
     *
     * @param depth          node depth
     * @param order          node order
     * @param name           name
     * @param sourcePosition source position
     *
     * @return NodeModel
     *
     * @see Name
     */
    public static NodeModel fromJavaxLangModelElementName(int depth, int order, Name name,
            PositionWithinSourceCode sourcePosition) {

        boolean isLeaf = true;
        String treeKind = NAME_TREE_KIND;
        String textualRepresentation = name.toString();

        Set<NodeModelProperty> emptyNodeModelProperties = Collections.emptySet();

        // null for javadoc => Name does not have javadoc
        return createNodeModel(depth, order, isLeaf, treeKind, textualRepresentation, null, sourcePosition,
                                                                                            emptyNodeModelProperties);
    }

    /**
     * Creates NodeModel from javax.lang.model.element.Modifier type. Modifier does not have properties and is always
     * leaf.
     *
     * @param depth          node depth
     * @param order          node order
     * @param modifier       modifier
     * @param sourcePosition source position
     *
     * @return NodeModel
     *
     * @see Modifier
     */
    public static NodeModel fromJavaxLangModelElementModifier(int depth, int order, Modifier modifier,
            PositionWithinSourceCode sourcePosition) {

        boolean isLeaf = true;
        String treeKind = MODIFIER_TREE_KIND;
        String textualRepresentation = modifier.toString();

        Set<NodeModelProperty> emptyNodeModelProperties = Collections.emptySet();

        // null for javadoc => Modifier does not have javadoc
        return createNodeModel(depth, order, isLeaf, treeKind, textualRepresentation, null, sourcePosition,
                                                                                            emptyNodeModelProperties);
    }

    /**
     * Creates NodeModel from javax.lang.model.type.TypeKind type. TypeKind does not have properties and is always leaf.
     *
     * @param depth          node depth
     * @param order          node order
     * @param typeKind       type kind
     * @param sourcePosition source position
     *
     * @return NodeModel
     *
     * @see TypeKind
     */
    public static NodeModel fromJavaxLangModelTypeTypeKind(int depth, int order, TypeKind typeKind,
            PositionWithinSourceCode sourcePosition) {

        boolean isLeaf = true;
        String treeKind = TYPE_KIND_TREE_KIND;
        String textualRepresentation = typeKind.toString().toLowerCase();

        Set<NodeModelProperty> emptyNodeModelProperties = Collections.emptySet();

        // null for javadoc => TypeKind does not have javadoc
        return createNodeModel(depth, order, isLeaf, treeKind, textualRepresentation, null, sourcePosition,
                                                                                            emptyNodeModelProperties);
    }


    //-----------------------------------------------------------------------------------------------------------------
    // Private Helpers.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Creates NodeModel according to given arguments.
     *
     * @param depth                    depth
     * @param order                    order
     * @param isLeaf                   leaf flag
     * @param treeKind                 tree kind
     * @param textualRepresentation    textual representation
     * @param javadoc                  javadoc
     * @param positionWithinSourceCode position within source code
     * @param nodeModelProperties      node model properties
     *
     * @return NodeModel
     */
    private static NodeModel createNodeModel(int depth, int order, boolean isLeaf, String treeKind,
            String textualRepresentation, String javadoc, PositionWithinSourceCode positionWithinSourceCode,
            Set<NodeModelProperty> nodeModelProperties) {
        return new NodeModel(depth,
                             order,
                             isLeaf,
                             treeKind,
                             textualRepresentation,
                             javadoc,
                             positionWithinSourceCode,
                             nodeModelProperties);
    }

    /**
     * Determines if given tree is leaf.
     *
     * @param tree tree
     *
     * @return <b>true</b>  if given tree is leaf<br />
     *         <b>false</b> otherwise
     */
    private static boolean determineIfTreeIsLeaf(Tree tree) {
        if (tree.getKind() == Tree.Kind.BLOCK) {
            boolean blockHasNoStatements = ((BlockTree) tree).getStatements().isEmpty();

            return blockHasNoStatements; // no statements => is leaf
        }

        return false; // most of trees will not be leafs
    }
}
