/*
 * Copyright 2012 Martin Skurla
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.crazyjavahacking.astvisualizer.impl.javac;


import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.lang.model.element.Modifier;

import org.crazyjavahacking.astvisualizer.impl.javac.model.EdgeDataFactory;
import org.crazyjavahacking.astvisualizer.impl.javac.model.IEdgeData;
import org.crazyjavahacking.astvisualizer.impl.javac.util.CollectionUtils;

import com.sun.source.tree.AnnotationTree;
import com.sun.source.tree.AssertTree;
import com.sun.source.tree.BlockTree;
import com.sun.source.tree.BreakTree;
import com.sun.source.tree.CaseTree;
import com.sun.source.tree.ClassTree;
import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.ContinueTree;
import com.sun.source.tree.ErroneousTree;
import com.sun.source.tree.ForLoopTree;
import com.sun.source.tree.IdentifierTree;
import com.sun.source.tree.IfTree;
import com.sun.source.tree.LabeledStatementTree;
import com.sun.source.tree.MemberSelectTree;
import com.sun.source.tree.MethodInvocationTree;
import com.sun.source.tree.MethodTree;
import com.sun.source.tree.ModifiersTree;
import com.sun.source.tree.NewArrayTree;
import com.sun.source.tree.NewClassTree;
import com.sun.source.tree.ParameterizedTypeTree;
import com.sun.source.tree.PrimitiveTypeTree;
import com.sun.source.tree.ReturnTree;
import com.sun.source.tree.SwitchTree;
import com.sun.source.tree.Tree;
import com.sun.source.tree.TryTree;
import com.sun.source.tree.TypeParameterTree;
import com.sun.source.tree.UnionTypeTree;
import com.sun.source.tree.VariableTree;
import com.sun.source.tree.WildcardTree;
import com.sun.tools.javac.tree.JCTree;

import static org.crazyjavahacking.astvisualizer.impl.javac.ASTTransitionNames.*;


/**
 * Represents all processed Tree types.
 *
 * @author Martin Skurla (crazyjavahacking@gmail.com)
 */
enum TreeTypeEnum {

    //-----------------------------------------------------------------------------------------------------------------
    // Constants.
    //-----------------------------------------------------------------------------------------------------------------
    ANNOTATION_TREE(JCTree.JCAnnotation.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            AnnotationTree annotationTree = (AnnotationTree) tree;
            final int numberOfArguments = annotationTree.getArguments().size();

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>(numberOfArguments + 1); // 1 for annotation type
            edgeData.add   (EdgeDataFactory.fromTree(ANNOTATION_TYPE));
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(ARGUMENT, numberOfArguments));

            return edgeData;
        }
    },
    ARRAY_ACCESS_TREE(2, JCTree.JCArrayAccess.class,   EXPRESSION, INDEX),
    ARRAY_TYPE_TREE  (1, JCTree.JCArrayTypeTree.class, TYPE),
    ASSERT_TREE(JCTree.JCAssert.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            AssertTree assertTree = (AssertTree) tree;

            return assertTree.getDetail() == null ? Collections.singletonList(EdgeDataFactory.fromTree(CONDITION))
                                                  : CollectionUtils.listOf(EdgeDataFactory.fromTree(CONDITION),
                                                                           EdgeDataFactory.fromTree(DETAIL));
        }
    },
    ASSIGNMENT_TREE(2, JCTree.JCAssign.class, VARIABLE,     EXPRESSION),
    BINARY_TREE    (2, JCTree.JCBinary.class, LEFT_OPERAND, RIGHT_OPERAND),
    BLOCK_TREE(JCTree.JCBlock.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            BlockTree blockTree = (BlockTree) tree;

            return EdgeDataFactory.fromTreeMultipleTimes(STATEMENT, blockTree.getStatements().size());
        }
    },
    BREAK_TREE(JCTree.JCBreak.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            BreakTree breakTree = (BreakTree) tree;

            return breakTree.getLabel() == null
                    ? Collections.<IEdgeData>emptyList()
                    : Collections.singletonList(EdgeDataFactory.fromName(LABEL, breakTree.getLabel()));
        }
    },
    CASE_TREE(JCTree.JCCase.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            CaseTree caseTree = (CaseTree) tree;
            final int numberOfStatements = caseTree.getStatements().size();

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>(numberOfStatements + 1); // for expression
            edgeData.add   (EdgeDataFactory.fromTree(EXPRESSION));
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(STATEMENT, numberOfStatements));

            return edgeData;
        }
    },
    CATCH_TREE(2, JCTree.JCCatch.class, PARAMETER, BLOCK),
    CLASS_TREE(JCTree.JCClassDecl.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            ClassTree classTree = (ClassTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            if (classTree.getModifiers() != null) {
                ModifiersTree modifiersTree = classTree.getModifiers();

                if (ModifiersTreeHelper.hasFlagsOrAnnotations(modifiersTree)) {
                    edgeData.add(EdgeDataFactory.fromTree(MODIFIERS));
                }
            }

            edgeData.add(EdgeDataFactory.fromName(SIMPLE_NAME, classTree.getSimpleName()));

            int numberOfTypeParameters = classTree.getTypeParameters().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(TYPE_PARAMETER, numberOfTypeParameters));

            if (classTree.getExtendsClause() != null) {
                edgeData.add(EdgeDataFactory.fromTree(EXTENDS_CLAUSE));
            }

            int numberOfImplementClauses = classTree.getImplementsClause().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(IMPLEMENTS_CLAUSE, numberOfImplementClauses));

            int numberOfMembers = classTree.getMembers().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(MEMBER, numberOfMembers));

            return edgeData;
        }
    },
    COMPILATION_UNIT_TREE(JCTree.JCCompilationUnit.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            CompilationUnitTree compilationUnitTree = (CompilationUnitTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            // just package could have package annotations
            int numberOfPackageAnnotations = compilationUnitTree.getPackageAnnotations().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(PACKAGE_ANNOTATION, numberOfPackageAnnotations));

            // class could have package declaration, package has to have package declaration
            if (compilationUnitTree.getPackageName() != null) {
                edgeData.add(EdgeDataFactory.fromTree(PACKAGE_NAME));
            }

            // both class and package could have import
            int numberOfImports = compilationUnitTree.getImports().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(IMPORT, numberOfImports));

            // just class could have type declarations
            int numberOfTypeDeclarations = compilationUnitTree.getTypeDecls().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(TYPE_DECLARATION, numberOfTypeDeclarations));

            return edgeData;
        }
    },
    COMPOUND_ASSIGNMENT_TREE   (2, JCTree.JCAssignOp.class,    VARIABLE,  EXPRESSION),
    CONDITIONAL_EXPRESSION_TREE(3, JCTree.JCConditional.class, CONDITION, TRUE_EXPRESSION, FALSE_EXPRESSION),
    CONTINUE_TREE(JCTree.JCContinue.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            ContinueTree continueTree = (ContinueTree) tree;

            return continueTree.getLabel() == null
                    ? Collections.<IEdgeData>emptyList()
                    : Collections.singletonList(EdgeDataFactory.fromName(LABEL, continueTree.getLabel()));
        }
    },
    DO_WHILE_LOOP_TREE    (2, JCTree.JCDoWhileLoop.class,     STATEMENT, EXPRESSION),
    EMPTY_STATEMENT_TREE  (0, JCTree.JCSkip.class),
    ENHANCED_FOR_LOOP_TREE(3, JCTree.JCEnhancedForLoop.class, VARIABLE,  EXPRESSION, STATEMENT),
    ERRONEOUS_TREE(JCTree.JCErroneous.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            ErroneousTree erroneousTree = (ErroneousTree) tree;

            return EdgeDataFactory.fromTreeMultipleTimes(ERROR_TREE, erroneousTree.getErrorTrees().size());
        }
    },
    EXPRESSION_STATEMENT_TREE(1, JCTree.JCExpressionStatement.class, EXPRESSION),
    FOR_LOOP_TREE(JCTree.JCForLoop.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            ForLoopTree forLoopTree = (ForLoopTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            int numberOfInitializers = forLoopTree.getInitializer().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(INITIALIZER, numberOfInitializers));

            if (forLoopTree.getCondition() != null) {
                edgeData.add(EdgeDataFactory.fromTree(CONDITION));
            }

            int numberOfUpdates = forLoopTree.getUpdate().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(UPDATE, numberOfUpdates));

            edgeData.add(EdgeDataFactory.fromTree(STATEMENT));

            return edgeData;
        }
    },
    IDENTIFIER_TREE(JCTree.JCIdent.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            IdentifierTree identifierTree = (IdentifierTree) tree;

            return Collections.singletonList(EdgeDataFactory.fromName(NAME, identifierTree.getName()));
        }
    },
    IF_TREE(JCTree.JCIf.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            IfTree ifTree = (IfTree) tree;

            return ifTree.getElseStatement() == null ? CollectionUtils.listOf(EdgeDataFactory.fromTree(CONDITION),
                                                                              EdgeDataFactory.fromTree(THEN_STATEMENT))
                                                     : CollectionUtils.listOf(EdgeDataFactory.fromTree(CONDITION),
                                                                              EdgeDataFactory.fromTree(THEN_STATEMENT),
                                                                              EdgeDataFactory.fromTree(ELSE_STATEMENT));
        }
    },
    IMPORT_TREE     (1, JCTree.JCImport.class,     QUALIFIED_IDENTIFIER),
    INSTANCE_OF_TREE(2, JCTree.JCInstanceOf.class, EXPRESSION, TYPE),
    LABELED_STATEMENT_TREE(JCTree.JCLabeledStatement.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            LabeledStatementTree labeledStatementTree = (LabeledStatementTree) tree;

            return CollectionUtils.listOf(EdgeDataFactory.fromName(LABEL, labeledStatementTree.getLabel()),
                                          EdgeDataFactory.fromTree(STATEMENT));
        }
    },
    LITERAL_TREE(0, JCTree.JCLiteral.class),
    MEMBER_SELECT_TREE(JCTree.JCFieldAccess.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            MemberSelectTree memberSelectTree = (MemberSelectTree) tree;

            return CollectionUtils.listOf(EdgeDataFactory.fromTree(EXPRESSION),
                                          EdgeDataFactory.fromName(IDENTIFIER, memberSelectTree.getIdentifier()));
        }
    },
    METHOD_INVOCATION_TREE(JCTree.JCMethodInvocation.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            MethodInvocationTree methodInvocationTree = (MethodInvocationTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            int numberOfTypeArguments = methodInvocationTree.getTypeArguments().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(TYPE_ARGUMENT, numberOfTypeArguments));

            edgeData.add(EdgeDataFactory.fromTree(METHOD_SELECT));

            int numberOfArguments = methodInvocationTree.getArguments().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(ARGUMENT, numberOfArguments));

            return edgeData;
        }
    },
    METHOD_TREE(JCTree.JCMethodDecl.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            MethodTree methodTree = (MethodTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            // all type of methods (class, interface, annotation) has modifiers + type + name
            if (methodTree.getModifiers() != null) {
                ModifiersTree modifiersTree = methodTree.getModifiers();

                if (ModifiersTreeHelper.hasFlagsOrAnnotations(modifiersTree)) {
                    edgeData.add(EdgeDataFactory.fromTree(MODIFIERS));
                }
            }

            // constructor does not have return type
            if (methodTree.getReturnType() != null) {
                edgeData.add(EdgeDataFactory.fromTree(RETURN_TYPE));
            }

            edgeData.add(EdgeDataFactory.fromName(NAME, methodTree.getName()));

            // class + interface methods could have type parameters
            int numberOfTypeParameters = methodTree.getTypeParameters().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(TYPE_PARAMETER, numberOfTypeParameters));

            // class + interface methods could have parameters
            int numberOfParameters = methodTree.getParameters().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(PARAMETER, numberOfParameters));

            // class + interface methods could have throw clause
            int numberOfThrows = methodTree.getThrows().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(THROW, numberOfThrows));

            // class method has body, but interface and annotation do not
            if (methodTree.getBody() != null) {
                edgeData.add(EdgeDataFactory.fromTree(BODY));
            }

            // annotation method could have optional default return value
            if (methodTree.getDefaultValue() != null) {
                edgeData.add(EdgeDataFactory.fromTree(DEFAULT_VALUE));
            }

            return edgeData;
        }
    },
    MODIFIERS_TREE(JCTree.JCModifiers.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            ModifiersTree modifiersTree = (ModifiersTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            for (Modifier modifier : modifiersTree.getFlags()) {
                edgeData.add(EdgeDataFactory.fromModifier(FLAG, modifier));
            }

            int numberOfAnnotations = modifiersTree.getAnnotations().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(ANNOTATION, numberOfAnnotations));

            return edgeData;
        }
    },
    NEW_ARRAY_TREE(JCTree.JCNewArray.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            NewArrayTree newArrayTree = (NewArrayTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            // not every array has to have type, for example array initializer => Object[] array = {"", ""};
            if (newArrayTree.getType() != null) {
                edgeData.add(EdgeDataFactory.fromTree(TYPE));
            }

            int numberOfDimensions = newArrayTree.getDimensions().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(DIMENSION, numberOfDimensions));

            if (newArrayTree.getInitializers() != null) {
                int numberOfInitializers = newArrayTree.getInitializers().size();
                edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(INITIALIZER, numberOfInitializers));
            }

            return edgeData;
        }
    },
    NEW_CLASS_TREE(JCTree.JCNewClass.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            NewClassTree newClassTree = (NewClassTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            if (newClassTree.getEnclosingExpression() != null) {
                edgeData.add(EdgeDataFactory.fromTree(ENCLOSING_EXPRESSION));
            }

            edgeData.add(EdgeDataFactory.fromTree(IDENTIFIER));

            int numberOfTypeArguments = newClassTree.getTypeArguments().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(TYPE_ARGUMENT, numberOfTypeArguments));

            int numberOfArguments = newClassTree.getArguments().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(ARGUMENT, numberOfArguments));

            if (newClassTree.getClassBody() != null) {
                edgeData.add(EdgeDataFactory.fromTree(CLASS_BODY));
            }

            return edgeData;
        }
    },
    PARAMETERIZED_TYPE_TREE(JCTree.JCTypeApply.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            ParameterizedTypeTree parameterizedTypeTree = (ParameterizedTypeTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            edgeData.add(EdgeDataFactory.fromTree(TYPE));

            int numberOfTypeArguments = parameterizedTypeTree.getTypeArguments().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(TYPE_ARGUMENT, numberOfTypeArguments));

            return edgeData;
        }
    },
    PARENTHESIZED_TREE(1, JCTree.JCParens.class, EXPRESSION),
    PRIMITIVE_TYPE_TREE(JCTree.JCPrimitiveTypeTree.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            PrimitiveTypeTree primitiveTypeTree = (PrimitiveTypeTree) tree;

            return Collections.singletonList(EdgeDataFactory.fromTypeKind(PRIMITIVE_TYPE_KIND,
                                                                          primitiveTypeTree.getPrimitiveTypeKind()));
        }
    },
    RETURN_TREE(JCTree.JCReturn.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
             ReturnTree returnTree = (ReturnTree) tree;

            return returnTree.getExpression() == null ? Collections.<IEdgeData>emptyList()
                                                      : Collections.singletonList(EdgeDataFactory.fromTree(EXPRESSION));
        }
    },
    SWITCH_TREE(JCTree.JCSwitch.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            SwitchTree switchTree = (SwitchTree) tree;
            final int numberOfCases = switchTree.getCases().size();

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>(numberOfCases + 1); // 1 for expression
            edgeData.add(EdgeDataFactory.fromTree(EXPRESSION));
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(CASE, numberOfCases));

            return edgeData;
        }
    },
    SYNCHRONIZED_TREE(2, JCTree.JCSynchronized.class, EXPRESSION, BLOCK),
    THROW_TREE       (1, JCTree.JCThrow.class,        EXPRESSION),
    TRY_TREE(JCTree.JCTry.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            TryTree tryTree = (TryTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            // starting from Java 7
            int numberOfResources = tryTree.getResources().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(RESOURCE, numberOfResources));

            edgeData.add(EdgeDataFactory.fromTree(BLOCK));

            int numberOfCatches = tryTree.getCatches().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(CATCH, numberOfCatches));

            if (tryTree.getFinallyBlock() != null) {
                edgeData.add(EdgeDataFactory.fromTree(FINALLY_BLOCK));
            }

            return edgeData;
        }
    },
    TYPE_CAST_TREE(2, JCTree.JCTypeCast.class, TYPE, EXPRESSION),
    TYPE_PARAMETER_TREE(JCTree.JCTypeParameter.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            TypeParameterTree typeParameterTree = (TypeParameterTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();
            edgeData.add(EdgeDataFactory.fromName(NAME, typeParameterTree.getName()));

            int numberOfBounds = typeParameterTree.getBounds().size();
            edgeData.addAll(EdgeDataFactory.fromTreeMultipleTimes(BOUND, numberOfBounds));

            return edgeData;
        }
    },
    UNARY_TREE(1, JCTree.JCUnary.class, EXPRESSION),

    // starting from Java 7
    UNION_TYPE_TREE(JCTree.JCTypeUnion.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            UnionTypeTree unionTypeTree = (UnionTypeTree) tree;

            int numberOfTypeAlternatives = unionTypeTree.getTypeAlternatives().size();

            return EdgeDataFactory.fromTreeMultipleTimes(TYPE_ALTERNATIVE, numberOfTypeAlternatives);
        }
    },
    VARIABLE_TREE(JCTree.JCVariableDecl.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            VariableTree variableTree = (VariableTree) tree;

            List<IEdgeData> edgeData = new ArrayList<IEdgeData>();

            if (variableTree.getModifiers() != null) {
                ModifiersTree modifiersTree = variableTree.getModifiers();

                if (ModifiersTreeHelper.hasFlagsOrAnnotations(modifiersTree)) {
                    edgeData.add(EdgeDataFactory.fromTree(MODIFIERS));
                }
            }

            edgeData.add(EdgeDataFactory.fromTree(TYPE));
            edgeData.add(EdgeDataFactory.fromName(NAME, variableTree.getName()));

            if (variableTree.getInitializer() != null) {
                edgeData.add(EdgeDataFactory.fromTree(INITIALIZER));
            }

            return edgeData;
        }
    },
    WHILE_LOOP_TREE(2, JCTree.JCWhileLoop.class, CONDITION, STATEMENT),
    WILDCARD_TREE(JCTree.JCWildcard.class) {
        @Override
        List<IEdgeData> edgeDataInternal(Tree tree) {
            WildcardTree wildcardTree = (WildcardTree) tree;

            return wildcardTree.getBound() == null ? Collections.<IEdgeData>emptyList()
                                                   : Collections.singletonList(EdgeDataFactory.fromTree(BOUND));
        }
    };


    //-----------------------------------------------------------------------------------------------------------------
    // Members.
    //-----------------------------------------------------------------------------------------------------------------
    private final List<IEdgeData> edgeData;
    private final Class<? extends JCTree> treeTypeClass;


    //-----------------------------------------------------------------------------------------------------------------
    // Constructors.
    //-----------------------------------------------------------------------------------------------------------------
    private TreeTypeEnum(Class<? extends JCTree> treeTypeClass) {
        this.treeTypeClass = treeTypeClass;
        this.edgeData = null;
    }

    private TreeTypeEnum(int edges, Class<? extends JCTree> treeTypeClass, String... transitionNames) {
        if (edges != transitionNames.length) {
            throw new IllegalArgumentException("For TypeTree: " + this);
        }

        this.treeTypeClass = treeTypeClass;

        List<IEdgeData> edgeDataList = new ArrayList<IEdgeData>(transitionNames.length);
        for (String transitionName : transitionNames) {
            edgeDataList.add(EdgeDataFactory.fromTree(transitionName));
        }
        this.edgeData = edgeDataList;
    }


    //-----------------------------------------------------------------------------------------------------------------
    // Public Methods.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Returns unmodifiable list of IEdgeData objects.
     *
     * @param tree tree
     *
     * @return unmodifiable list of IEdgeData objects
     */
    public final List<IEdgeData> getEdgeData(Tree tree) {
        return Collections.unmodifiableList(edgeDataInternal(tree));
    }

    /**
     * Gets instance by given JCTree type class.
     *
     * @param treeTypeClass JCTree type class
     *
     * @return TreeTypeEnum instance
     *
     * @throws IllegalArgumentException if the given type is not supported
     */
    public static TreeTypeEnum getByClass(Class<? extends JCTree> treeTypeClass) {
        for (TreeTypeEnum treeType : values()) {
            if (treeType.treeTypeClass == treeTypeClass) {
                return treeType;
            }
        }

        throw new IllegalArgumentException("Given type is not supported: " + treeTypeClass);
    }


    //-----------------------------------------------------------------------------------------------------------------
    // Private Helpers.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Returns list of IEdgeData objects.
     *
     * @param tree tree
     *
     * @return list of IEdgeData objects
     */
    List<IEdgeData> edgeDataInternal(Tree tree) {
        return edgeData;
    }
}
